unit Univbm;
{

  Universal BitMap Loader v1.0b by Daniel Parnell

  Mail to:
    s921878@minyos.its.rmit.edu.au

  Or WEB page:
    http://minyos.its.rmit.edu.au/~s921878

  Images that can be loaded with this component are limited to the screen's
  physical size (but not colour depth).  If you attempt to load a bitmap that
  is bigger than the screen, only that portion of the bitmap that will fit on
  the screen is loaded correctly.

}

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs;

type
  TBuffer = array[0..1023] of byte;
  PBuffer = ^TBuffer;

  TImageBuffer=array[0..$fff0] of byte;
  PImageBuffer=^TImageBuffer;

  EBMException = class(exception);

  TBMLoader = class
    FormatName,
    FormatMask  : string;
    hPal        : HPALETTE;
    pal         : PLogPalette;
    form        : TForm;
    infile      : file;
    buffer      : PBuffer;
    pos         : word;

    constructor Create(aName,aMask : string); virtual;
    procedure   Load(name : string; bitmap : TBitmap); virtual; abstract;
    procedure   StartPalette(colours : integer);
    procedure   SetPalette(c,r,g,b : byte);
    procedure   FinishPalette(colours:integer);
    procedure   CleanUp;
    procedure   BeginBitmap(wide,high:word);
    procedure   WriteLine(wide,line : word; bits : byte; bitData : pointer);
    procedure   BitmapDone(bitmap:TBitmap; wide,high : word);
    function    NextByte:byte;
    procedure   FileOpen(name : string);
    procedure   FileSeek(where : longint);
    procedure   FileClose;
  end;

  TUnivBMLoader = class(TComponent)
  private
    { Private declarations }
    fBitmap  : TBitmap;

    function GetFormats:string;
  protected
    { Protected declarations }
  public
    { Public declarations }
    constructor Create(AOwner : TComponent); override;
    destructor  Destroy; override;
    procedure   LoadFromFile(name : string);
    function    FindBMType(mask : string):integer;
  published
    { Published declarations }
    property Formats : string  read GetFormats;
    property Bitmap  : TBitmap read fBitmap;
  end;

procedure Register;

implementation

type
  TBMP = class(TBMLoader)
    procedure Load(name : string; bitmap : TBitmap); override;
  end;

  TICN = class(TBMLoader)
    procedure Load(name : string; bitmap : TBitmap); override;
  end;

  TWMF = class(TBMLoader)
    procedure Load(name : string; bitmap : TBitmap); override;
  end;

var
  fFormats    : TStringList;
  OldExitProc : pointer;

function GetExt(s : string):string;
begin
  result:=copy(s,pos('|',s)+3,255);
end;

procedure RegisterBMLoader(name,mask : string; loader : TBMLoader);
var
  index : integer;

begin
  index:=fFormats.Add(name+'|*.'+mask);
  fFormats.Objects[index]:=loader;
end;

constructor TUnivBMLoader.Create(AOwner : TComponent);
begin
  inherited Create(AOwner);

  fBitmap:=TBitmap.Create;
end;

destructor TUnivBMLoader.Destroy;
begin
  fBitmap.Free;
  inherited Destroy;
end;

function TUnivBMLoader.FindBMType(mask : string):integer;
var
  i : integer;

begin
  result:=-1;
  mask:=UpperCase(mask);

  for i:=0 to fFormats.Count-1 do
    if mask=GetExt(fFormats.Strings[i]) then
      result:=i;
end;

function TUnivBMLoader.GetFormats:string;
var
  i : integer;

begin
  result:='Any File|*.*|';
  for i:=0 to fFormats.Count-1 do
    result:=result+fFormats.Strings[i]+'|';

  result:=copy(result,1,length(result)-1);
end;

procedure TUnivBMLoader.LoadFromFile(name : string);
var
  index : integer;

begin
  index:=FindBMType(copy(ExtractFileExt(name),2,3));

  if index<>-1 then
    TBMLoader(fFormats.Objects[index]).Load(name,bitmap)
  else
    raise EBMException.Create('No load method for '+name);
end;

constructor TBMLoader.Create(aName,aMask : string);
begin
  inherited Create;

  FormatName:=aName;
  FormatMask:=UpperCase(aMask);
  RegisterBMLoader(FormatName,FormatMask,Self);
end;

procedure TBMLoader.StartPalette(colours : integer);
begin
  getmem(pal,4+colours*4);
{  fillchar(pal^,4+colours*4,0);}
  pal^.palVersion:=$300;
  pal^.palNumEntries:=256;
end;

procedure TBMLoader.SetPalette(c,r,g,b : byte);
begin
{$R-}
  pal^.palpalentry[c].pered:=r;
  pal^.palpalentry[c].pegreen:=g;
  pal^.palpalentry[c].peblue:=b;
  pal^.palpalentry[c].peflags:=0;
{$R+}
end;

procedure TBMLoader.FinishPalette(colours:integer);
begin
  hPal:=createpalette(pal^);
  freemem(pal,4+colours*4);
end;

procedure TBMLoader.BeginBitmap(wide,high:word);
begin
  form:=TForm.Create(nil);
  form.borderStyle:=bsDialog;
  form.caption:='Loading';
  form.clientWidth:=wide;
  form.clientHeight:=high;
  form.FormStyle:=fsStayOnTop;
  form.position:=poScreenCenter;
  form.Show;
end;

procedure TBMLoader.WriteLine(wide,line : word; bits : byte; bitData : pointer);
var
  BitsInfo : pbitmapinfo;
  i        : integer;
  oldpal   : hpalette;
  dc       : HDC;
  hbm      : HBITMAP;
  oldBM    : HBITMAP;
  screenDC : HDC;
  bbb      : WinTypes.TBITMAP;

begin
  getmem(bitsinfo,sizeof(tbitmapinfoheader)+256*4);
  with bitsinfo^ do
    begin
      with bmiheader do
        begin
          bisize:=sizeof(bmiHeader);
          biwidth:=wide;
          biheight:=1;
          biplanes:=1;
          bibitcount:=bits;
          bicompression:=BI_RGB;
          bisizeimage:=0;
          bixpelspermeter:=0;
          biypelspermeter:=0;
          biclrused:=0;
          biclrimportant:=0;
        end;
 {$R-}
      for i:=0 to (1 shl (bits-1)) do
        begin
          bmicolors[i].rgbblue:=2*i;
          bmicolors[i].rgbgreen:=0;
          bmicolors[i].rgbred:=2*i+1;
          bmicolors[i].rgbreserved:=0;
        end;
{$R+}
    end;

  dc:=form.canvas.handle;

  oldpal:=selectpalette(dc,hPal,False);
  RealizePalette(dc);
  i:=SetDIBitsToDevice(dc,0,line,wide,1,0,0,0,1,bitdata,BitsInfo^,DIB_PAL_COLORS);
  selectpalette(dc,oldpal,false);
  RealizePalette(dc);
  freemem(bitsinfo,sizeof(tbitmapinfoheader)+256*4);
end;

procedure TBMLoader.BitmapDone(bitmap:TBitmap; wide,high : word);
var
  dc       : HDC;
  hbm      : HBITMAP;
  screenDC : HDC;

begin
  dc:=form.canvas.handle;
  screenDC:=GetDC(0);
  hbm:=CreateCompatibleBitmap(screenDC,wide,high);
  releaseDC(0,screenDC);
  bitmap.handle:=hbm;
  bitmap.palette:=hPal;
  bitBlt(bitmap.canvas.handle,0,0,wide,high,dc,0,0,SRCCOPY);
  form.Free;
end;

function TBMLoader.NextByte:byte;
var
  read : word;

begin
  inc(pos);
  if pos>=1024 then
    begin
      pos:=0;
      blockRead(infile,buffer^,1024,read);
    end;

  result:=buffer^[pos];
end;

procedure TBMLoader.FileOpen(name : string);
begin
  AssignFile(infile,name);
  reset(infile,1);
  new(buffer);
  pos:=1024;
end;

procedure TBMLoader.FileSeek(where : longint);
begin
  pos:=1024;
  Seek(infile,where);
end;

procedure TBMLoader.FileClose;
begin
  CloseFile(infile);
  dispose(buffer);
end;

procedure TBMLoader.Cleanup;
begin
end;

procedure TBMP.Load(name : string; bitmap : TBitmap);
begin
  bitmap.LoadFromFile(name);
end;

procedure TICN.Load(name : string; bitmap : TBitmap);
var
  icon : TIcon;

begin
  icon:=TIcon.Create;
  try
    icon.LoadFromFile(name);
    bitmap.width:=icon.width;
    bitmap.height:=icon.height;
    bitmap.Canvas.Draw(0,0,icon);
  finally
    icon.Free;
  end;
end;

procedure TWMF.Load(name : string; bitmap : TBitmap);
var
  wmf : TMetafile;

begin
  wmf:=TMetaFile.Create;
  try
    wmf.LoadFromFile(name);
    BeginBitmap(wmf.width,wmf.height);
    Form.Canvas.Draw(0,0,wmf);
    BitmapDone(bitmap,wmf.width,wmf.height);
  finally
    wmf.Free;
  end;
end;

{$F+}
procedure MyExitProc;
var
  i : integer;

begin
  exitProc:=OldExitProc;

  for i:=0 to fFormats.Count-1 do
    TBMLoader(fFormats.Objects[i]).Free;

  fFormats.Free;
end;
{$F-}

procedure Register;
begin
  RegisterComponents('Bitmap', [TUnivBMLoader]);
end;

initialization
  oldExitProc:=ExitProc;
  ExitProc:=@MyExitProc;

  fFormats:=TStringList.Create;
  TBMP.Create('Windows Bitmap','BMP');
  TICN.Create('Windows Icon','ICO');
  TWMF.Create('Windows Metafile','WMF');
end.
